from abc import abstractmethod
from bibtexparser.middlewares.names import List
from pygls.protocol.language_server import LanguageServerProtocol

from bibli_ls.bibli_config import BackendConfig, BibliBibDatabase


class BibliBackend:
    backend_type: str
    _lsp: LanguageServerProtocol
    _config: BackendConfig

    def __init__(self, config: BackendConfig, lsp: LanguageServerProtocol):
        self._lsp = lsp
        self._config = config

    @abstractmethod
    def get_libraries(self) -> List[BibliBibDatabase]:
        """Return the libraries based on backend config"""
        pass



"""Test the CLI."""

import sys
from io import StringIO

from hamcrest import assert_that, is_

from bibli_ls import __version__
from bibli_ls.cli import cli


class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio  # free up some memory
        sys.stdout = self._stdout


def test_get_version():
    """Test that version is printed correctly"""

    sys.argv = ["bibli_ls", "--version"]
    with Capturing() as output:
        cli()

    print(output)
    assert_that(output[0], is_(str(__version__)))



from multiprocessing.pool import AsyncResult
import os
from bibtexparser.middlewares.names import List
from lsprotocol.types import (
    MessageType,
    WorkDoneProgressBegin,
    WorkDoneProgressEnd,
    WorkDoneProgressReport,
)
from pygls.progress import Progress
from pyzotero.zotero import Zotero
from bibtexparser.library import Library
from bibli_ls.backends.backend import BibliBackend
from bibli_ls.bibli_config import BackendConfig, BibliBibDatabase
from bibtexparser import bibtexparser

import multiprocessing


class ZoteroBackend(BibliBackend):
    _zot: Zotero

    def __init__(self, config: BackendConfig, lsp):
        super().__init__(config, lsp)
        if config.library_id == "":
            lsp.show_message("Library ID not specified", MessageType.Error)
            return

        if config.api_key is None:
            lsp.show_message("API key not specified", MessageType.Error)
            return
        lsp.show_message(
            f"Initializing zotero API connection library_id `{config.library_id}`, library_type `{config.library_type}`"
        )
        self._zot = Zotero(config.library_id, config.library_type, config.api_key)

    def get_libraries(
        self,
    ) -> List[BibliBibDatabase]:
        count = self._zot.count_items()
        loaded = 0
        limit = 10

        library = Library()
        pool = multiprocessing.Pool(16)

        progress = Progress(self._lsp)
        progress.create("bibli")
        progress.begin(
            "bibli",
            WorkDoneProgressBegin(
                title=f"Retriving online library from `{self._zot.library_id}`",
                message="libraries loaded",
            ),
        )

        results: List[AsyncResult] = []
        for i in range(0, count, limit):
            results.append(
                pool.apply_async(
                    self._zot.items,
                    kwds={"start": i, "limit": limit, "content": "bibtex"},
                )
            )

        for r in results:
            items = r.get()
            loaded += len(items)
            items_str = "\n".join(items)
            bibtexparser.parse_string(items_str, library=library)

            progress.report(
                "bibli",
                WorkDoneProgressReport(
                    message="libraries loaded",
                    percentage=int(loaded * 100 / count),
                ),
            )

        progress.end(
            "bibli",
            WorkDoneProgressEnd(message="Done"),
        )

        pool.close()
        pool.join()

        # Writing to file
        filename = f".zotero_api_{self._zot.library_id}.bib"
        root_path = self._lsp.workspace.root_path
        cache_file = None
        if root_path:
            cache_file = os.path.join(root_path, filename)
            self._lsp.show_message(f"Writing to bibfile to `{cache_file}`")
            bibtexparser.write_file(cache_file, library)

        return [BibliBibDatabase(library, cache_file)]



from dataclasses import dataclass, field
from bibtexparser.library import Library
from lsprotocol.types import MessageType
from pygls.protocol.language_server import LanguageServerProtocol


DEFAULT_HEADER_FORMAT = [
    "# `{entry_type}` {title}",
    "_{author}_",
    "───────────────────────────────────────────────────────────────────────────────────",
]
DEFAULT_FOOTER_FORMAT = [
    "───────────────────────────────────────────────────────────────────────────────────",
    "\n" + "\t" * 20 + "from `{bibfile}`",
]
DEFAULT_CITE_PREFIX = "@"
DEFAULT_CITE_REGEX_STR = r"@([A-Za-z_0-9]+)\b"

DEFAULT_WRAP = 80
DEFAULT_CHAR_LIMIT = 400
# cite_regex = re.compile(prefix + "[A-Za-z_0-9]*$")


@dataclass
class BibliBibDatabase:
    library: Library
    path: str | None


@dataclass
class DocFormatingConfig:
    """Config for formating a documentation string"""

    wrap: int = DEFAULT_WRAP
    character_limit: int = DEFAULT_CHAR_LIMIT
    format: str = "list"
    show_fields: list[str] = field(default_factory=lambda: [])
    header_format: list[str] | str = field(
        default_factory=lambda: DEFAULT_HEADER_FORMAT
    )
    footer_format: list[str] | str = field(
        default_factory=lambda: DEFAULT_FOOTER_FORMAT
    )


@dataclass
class CiteConfig:
    prefix: str = DEFAULT_CITE_PREFIX
    regex: str = DEFAULT_CITE_REGEX_STR


@dataclass
class HoverConfig:
    doc_format: DocFormatingConfig = field(default_factory=lambda: DocFormatingConfig())


@dataclass
class CompletionConfig:
    doc_format: DocFormatingConfig = field(default_factory=lambda: DocFormatingConfig())


@dataclass
class BackendConfig:
    backend_type: str = "bibfile"
    library_id: str = ""
    library_type: str = "user"
    api_key: str | None = None
    bibfiles: list[str] = field(default_factory=lambda: [])


# TODO: Is there a better way to do this?
EXPECTED_VALUES = {
    "backend_type": ["zotero_api", "bibfile"],
    "library_type": ["user", "group"],
    "doc_format.format": ["table", "list"],
}


@dataclass
class BibliTomlConfig:
    """Runtime configurations used by bibli in one place"""

    backend: dict[str, BackendConfig] = field(default_factory=lambda: {})
    bibfiles: list[str] = field(default_factory=lambda: [])
    hover: HoverConfig = field(default_factory=lambda: HoverConfig())
    completion: CompletionConfig = field(default_factory=lambda: CompletionConfig())
    cite: CiteConfig = field(default_factory=lambda: CiteConfig())

    def check_expected(self, field, value, lsp) -> bool:
        if value not in EXPECTED_VALUES[field]:
            lsp.show_message(
                f"Unexpected value in {field}: {value}",
                MessageType.Error,
            )
            return False
        return True

    def sanitize(self, lsp: LanguageServerProtocol) -> bool:
        valid = True
        for _, v in self.backend.items():
            valid |= self.check_expected("backend_type", v.backend_type, lsp)
            match v.backend_type:
                case "zotero_api":
                    valid |= self.check_expected("library_type", v.library_type, lsp)
                case "bibfile":
                    pass

        valid |= self.check_expected(
            "doc_format.format", self.hover.doc_format.format, lsp
        )

        valid |= self.check_expected(
            "doc_format.format", self.completion.doc_format.format, lsp
        )

        return valid
